"""Utility functions for handling batch and pointer representations."""

import torch


def get_batch_from_ptr(ptr):
    """Reconstruct batch indices (batch) from pointer (ptr).

    Parameters
    ----------
    ptr : torch.Tensor
        Pointer tensor indicating the start of each batch.
        Tensor of shape (B+1,) where B is the number of batches.

    Returns
    -------
    torch.Tensor
        A tensor where each element indicates the batch index for each item.
        Tensor of shape (N,) where N is the total number of items across all batches.
    """
    return torch.arange(len(ptr) - 1, device=ptr.device).repeat_interleave(
        ptr[1:] - ptr[:-1],
    )


def get_ptr_from_batch(batch):
    """Reconstruct pointer (ptr) from batch indices (batch).

    Parameters
    ----------
    batch : torch.Tensor
        A tensor where each element indicates the batch index for each item.
        Tensor of shape (N,) where N is the total number of items across all batches.

    Returns
    -------
    torch.Tensor
        A pointer tensor indicating the start of each batch.
        Tensor of shape (B+1,) where B is the number of batches.
    """
    ptr = torch.cat(
        [
            torch.tensor([0], device=batch.device),
            torch.where(batch[1:] - batch[:-1] != 0)[0] + 1,
            torch.tensor([batch.shape[0]], device=batch.device),
        ],
        0,
    )
    return ptr


def get_edge_index_from_ptr(ptr, shape, remove_self_loops=True):
    """Construct edge index of fully connected graph from pointer (ptr).
    This function should be used for graphs represented by sparse tensors,
    i.e. graphs where the number of nodes per graph can vary.

    Parameters
    ----------
    ptr : torch.Tensor
        Pointer tensor indicating the start of each batch.
        Tensor of shape (B+1,) where B is the number of batches.
    shape : torch.Size
        Shape of the node tensor, expected to be (N, C)
    remove_self_loops : bool, optional
        Whether to remove self-loops from the edge index, by default True.

    Returns
    -------
    torch.Tensor
        A tensor of shape (2, E) where E is the number of edges, representing the edge index.
    """
    N = shape[0]

    diff = ptr[1:] - ptr[:-1]
    starts = torch.zeros(N, dtype=torch.long, device=ptr.device)
    starts.scatter_(0, ptr[:-1], torch.ones_like(ptr[:-1]))
    node2graph = starts.cumsum(0) - 1

    counts = diff[node2graph] - (1 if remove_self_loops else 0)
    E = counts.sum()
    row = torch.repeat_interleave(counts, output_size=E)

    g = node2graph[row]
    offset = ptr[g]
    row_local = row - offset

    start = counts.cumsum(0) - counts
    idx = torch.arange(row.numel(), device=ptr.device, dtype=torch.long)
    pos = idx - start[row]

    if remove_self_loops:
        col_local = pos + (pos >= row_local).to(torch.long)
    else:
        col_local = pos
    col = col_local + offset

    edge_index = torch.stack([row, col], dim=0)
    return edge_index


def get_edge_index_from_shape(shape, device=None, remove_self_loops=True):
    """Construct edge index of fully connected graph from the shape of a corresponding dense tensor.
    This function should be used for graphs represented by dense tensors,
    i.e. graphs where the number of nodes per graph is fixed.

    Parameters
    ----------
    shape : torch.Size
        Shape of the dense node tensor, expected to be (B, N, C)
        where B is the batch size and N is the number of nodes.
    device : torch.device
        Device on which the tensors are allocated. Default is CPU.
    remove_self_loops : bool, optional
        Whether to remove self-loops from the edge index, by default True.

    Returns
    -------
    torch.Tensor
        A tensor of shape (2, E) where E is the number of edges, representing the edge index.
    """
    B, N, _ = shape
    if device is None:
        device = torch.device("cpu")

    nodes = torch.arange(N, device=device)

    if remove_self_loops:
        row = nodes.repeat_interleave(N - 1)
        base = torch.arange(N - 1, device=device)
        i = nodes.unsqueeze(1)
        col_2d = base.unsqueeze(0).expand(N, -1)
        col_2d = col_2d + (col_2d >= i)
        col = col_2d.reshape(-1)
    else:
        row = nodes.repeat_interleave(N)
        col = nodes.repeat(N)

    edge_base = torch.stack([row, col], dim=0)

    offsets = torch.arange(B, device=device, dtype=torch.long) * N
    batched = edge_base.unsqueeze(2) + offsets.view(1, 1, -1)
    edge_index_global = batched.permute(0, 2, 1).reshape(2, -1)

    batch = torch.arange(B, device=device).repeat_interleave(N)
    return edge_index_global, batch


def get_edge_attr(fourmomenta, edge_index, eps=1e-10):
    """Calculate edge attributes based on the squared Lorentz norm of the sum of four-momenta.

    Parameters
    ----------
    fourmomenta : torch.Tensor
        A tensor of shape (B, N, 4) representing the four-momenta of particles.
    edge_index : torch.Tensor
        A tensor of shape (2, E) representing the edge index of the graph.
    eps : float, optional
        A small value to avoid log(0) issues, by default 1e-10.

    Returns
    -------
    torch.Tensor
        A tensor of shape (E,) representing the edge attributes, which are the logarithm of the squared Lorentz norm.
    """
    mij2 = lorentz_squarednorm(fourmomenta[edge_index[0]] + fourmomenta[edge_index[1]])
    edge_attr = mij2.clamp(min=eps).log()
    return edge_attr


def lorentz_squarednorm(v):
    """Lorentz norm, i.e. v^T @ g @ v

    Parameters
    ----------
    v : torch.Tensor
        Tensor of shape (..., 4)

    Returns
    -------
    torch.Tensor
        Lorentz norm of shape (..., )
    """
    t = v[..., 0] * v[..., 0]
    s = (v[..., 1:] * v[..., 1:]).sum(dim=-1)
    return t - s



import torch

from pelican.utils import get_batch_from_ptr, get_edge_index_from_ptr


def generate_batch(G=8, N_min=10, N_max=20, C=16, batch=None, edge_index=None):
    """
    Generate random batch, edge_index, graph, nodes, and edges for testing,
    assuming a fully connected graph.

    Parameters
    ----------
    G : int
        Number of graphs in the batch.
    N_range : list
        Range of number of nodes per graph.
    C : int
        Number of channels (features) for graph, nodes, and edges.
    batch : torch.Tensor, optional
        Predefined batch tensor, by default None.
    edge_index : torch.Tensor, optional
        Predefined edge_index tensor, by default None.

    Returns
    -------
    batch : torch.Tensor
        Batch tensor of shape (N).
    edge_index : torch.Tensor
        Edge index tensor of shape (2, E).
    graph : torch.Tensor
        Graph-level features of shape (G, C).
    nodes : torch.Tensor
        Node-level features of shape (G, N).
    edges : torch.Tensor
        Edge-level features of shape (G, E).
    """
    if batch is None or edge_index is None:
        length = torch.randint(low=N_min, high=N_max, size=(G,))
        ptr = torch.zeros(G + 1, dtype=torch.long)
        ptr[1:] = torch.cumsum(length, dim=0)
        batch = get_batch_from_ptr(ptr)
        shape = torch.Size((batch.shape[0], C))
        edge_index = get_edge_index_from_ptr(ptr, shape, remove_self_loops=False)
    else:
        assert batch is not None and edge_index is not None
    graph = torch.randn(G, C) if C > 0 else None
    nodes = torch.randn(batch.numel(), C) if C > 0 else None
    edges = torch.randn(edge_index.size(1), C) if C > 0 else None
    return batch, edge_index, graph, nodes, edges


def permute_single_graph(edge_index, graph=None, nodes=None, edges=None, permutation=None):
    # this code assumes a single graph (G=1)
    N = edge_index.max().item() + 1
    if permutation is None:
        permutation = torch.randperm(N)

    permutation_inverse = torch.empty_like(permutation)
    permutation_inverse[permutation] = torch.arange(N)
    edge_index_perm = permutation_inverse[edge_index]
    key = edge_index_perm[0] * N + edge_index_perm[1]
    order = key.argsort()

    graph_perm = graph if graph is not None else None
    nodes_perm = nodes[permutation] if nodes is not None else None
    edges_perm = edges[order] if edges is not None else None
    return permutation, graph_perm, nodes_perm, edges_perm, edge_index_perm[:, order]



"""PELICAN network architecture."""

import torch
from torch import nn
from torch.utils.checkpoint import checkpoint

from .layers import (
    Aggregator0to2,
    Aggregator1to2,
    Aggregator2to0,
    Aggregator2to1,
    Aggregator2to2,
    PELICANBlock,
)


class PELICAN(nn.Module):
    """PELICAN network.

    PELICAN stands for Permutation Equivariant and Lorentz Invariant or Covariant Aggregator Network.
    PELICAN takes input of varying rank (0, 1, or 2), projects them onto rank 2 objects (edges),
    processes them with multiple PELICANBlock, and extracts output of varying rank (0, 1, or 2).
    """

    def __init__(
        self,
        in_channels_rank2: int,
        in_channels_rank1: int,
        in_channels_rank0: int,
        out_rank: int,
        out_channels: int,
        num_blocks: int,
        hidden_channels: int,
        increase_hidden_channels: float = 1.0,
        map_multipliers: bool = True,
        factorize: bool = False,
        activation: str = "leaky_relu",
        dropout_prob: float | None = None,
        aggr: str = "mean",
        compile: bool = False,
        checkpoint_blocks: bool = False,
    ):
        """
        Parameters
        ----------
        in_channels_rank2 : int
            Edge-level input features (rank 2). Can be zero.
        in_channels_rank1 : int
            Node-level input features (rank 1). Can be zero.
        in_channels_rank0 : int
            Graph-level input features (rank 0). Can be zero.
        out_rank : int
            Rank of the output features (0, 1, or 2).
        out_channels : int
            Number of output channels.
        num_blocks : int
            Number of PELICAN blocks.
        hidden_channels : int
            Number of hidden channels.
        increase_hidden_channels : float
            Factor to increase hidden channels in the feedforward network. Default is 1.0.
        map_multipliers : bool
            Whether to use learnable multipliers for each aggregation map. Default is True.
        factorize : bool
            Whether to use factorized linear layers in the feedforward network. Default is False.
        activation : str
            Activation function to use ('gelu', 'relu', 'leaky_relu', 'tanh', 'sigmoid', 'silu'), by default 'leaky_relu'.
        dropout_prob: float
            Dropout probability in the feedforward network, by default None (no dropout).
        aggr : str
            Aggregation method to use ('mean', 'sum', 'prod', 'amin', 'amax'), by default 'mean'.
        compile : bool
            Whether to compile the model with torch.compile. Default is False.
            Compiling the model leads to significant speedups on GPU, because the aggregation functions involve many small operations that otherwise require many individual kernel launches.
            It is recommended to run ``model = torch.compile(model, **kwargs)`` outside of the constructor, however we provide this option for convenience.
            Note: When compile=True, the model requires the num_graphs argument in the forward pass to avoid a graph break.
        checkpoint_blocks : bool
            Whether to use gradient checkpointing for PELICAN blocks to save memory. Default is False.
        """
        super().__init__()
        layer_kwargs = dict(factorize=factorize, map_multipliers=map_multipliers, aggr=aggr)

        # embed inputs into edge features
        self.in_aggregator_rank1 = (
            Aggregator1to2(
                in_channels=in_channels_rank1,
                out_channels=in_channels_rank1,
                **layer_kwargs,
            )
            if in_channels_rank1 > 0
            else None
        )
        self.in_aggregator_rank0 = (
            Aggregator0to2(
                in_channels=in_channels_rank0,
                out_channels=in_channels_rank0,
                **layer_kwargs,
            )
            if in_channels_rank0 > 0
            else None
        )

        in_channels = in_channels_rank2 + in_channels_rank1 + in_channels_rank0
        assert in_channels > 0
        self.in_aggregator_rank2 = Aggregator2to2(
            in_channels=in_channels,
            out_channels=hidden_channels,
            **layer_kwargs,
        )

        # process edge features
        self._checkpoint_blocks = checkpoint_blocks
        self.blocks = nn.ModuleList(
            [
                PELICANBlock(
                    hidden_channels=hidden_channels,
                    increase_hidden_channels=increase_hidden_channels,
                    activation=activation,
                    dropout_prob=dropout_prob,
                    **layer_kwargs,
                )
                for _ in range(num_blocks)
            ]
        )

        # extract outputs from edge features
        out_aggregator_class = {
            0: Aggregator2to0,
            1: Aggregator2to1,
            2: Aggregator2to2,
        }
        self.out_aggregator = out_aggregator_class[out_rank](
            in_channels=hidden_channels,
            out_channels=out_channels,
            **layer_kwargs,
        )

        self.compile = compile
        if compile:
            # ugly hack to make torch.compile convenient for users
            # the clean solution is model = torch.compile(model, **kwargs) outside of the constructor
            self.__class__ = torch.compile(
                self.__class__, dynamic=True, fullgraph=True, mode="default"
            )

    def forward(
        self,
        edge_index,
        batch,
        in_rank2=None,
        in_rank1=None,
        in_rank0=None,
        num_graphs: int | None = None,
    ):
        """Forward pass.

        Parameters
        ----------
        edge_index : torch.Tensor
            Edge index tensor of shape (2, E).
        batch : torch.Tensor
            Batch tensor of shape (N).
        in_rank2 : torch.Tensor
            Edge-level input features of shape (E, in_channels_rank2), by default None.
        in_rank1 : torch.Tensor
            Node-level input features of shape (N, in_channels_rank1), by default None.
        in_rank0 : torch.Tensor
            Graph-level input features of shape (G, in_channels_rank0), by default None.
        num_graphs : int
            The number of graphs G in the batch, also known as batch size.
            If None, it will be inferred from the batch tensor.
            Inferring this number from the batch tensor requires a GPU/CPU synchronization,
            which slows down the code when running on GPU.
            Currently, the code requires the num_graphs argument in case compile=True.

        Returns
        -------
        out : torch.Tensor
            Output features of shape (G, out_channels) for out_rank=0, (N, out_channels) for out_rank=1, or (E, out_channels) for out_rank=2.
        """
        if num_graphs is None:
            assert (
                not self.compile
            ), "num_graphs must be provided when model is compiled, otherwise the .item() call breaks the computational graph, slowing down the compiled code."
            num_graphs = batch[-1].item() + 1

        # embed inputs into edge features
        edges = [in_rank2] if in_rank2 is not None else []
        if in_rank1 is not None and self.in_aggregator_rank1 is not None:
            edges_fromrank1 = self.in_aggregator_rank1(in_rank1, edge_index, batch)
            edges.append(edges_fromrank1)
        if in_rank0 is not None and self.in_aggregator_rank0 is not None:
            edges_fromrank0 = self.in_aggregator_rank0(in_rank0, edge_index, batch)
            edges.append(edges_fromrank0)
        edges = torch.cat(edges, dim=-1)
        x = self.in_aggregator_rank2(edges, edge_index, batch)

        # process edge features
        for block in self.blocks:
            kwargs = dict(x=x, edge_index=edge_index, batch=batch)
            if self._checkpoint_blocks:
                x = checkpoint(block, use_reentrant=False, **kwargs)
            else:
                x = block(**kwargs)

        # extract outputs from edge features
        out = self.out_aggregator(x, edge_index=edge_index, batch=batch, G=num_graphs)
        return out



import pytest
import torch

from pelican.utils import (
    get_batch_from_ptr,
    get_edge_index_from_ptr,
    get_edge_index_from_shape,
    get_ptr_from_batch,
)


@pytest.mark.parametrize("B, N", [(1, 5), (4, 9)])
@pytest.mark.parametrize("remove_self_loops", [True, False])
def test_get_edge_index_tools(B, N, remove_self_loops, C=16):
    # test that the two get_edge_index functions give the same result
    tensor = torch.randn(B, N, C)
    ptr = torch.arange(B + 1) * N

    edge_index_from_shape, _ = get_edge_index_from_shape(
        tensor.shape, tensor.device, remove_self_loops=remove_self_loops
    )
    tensor_sparse = tensor.view(B * N, C)
    edge_index_from_ptr = get_edge_index_from_ptr(
        ptr, tensor_sparse.shape, remove_self_loops=remove_self_loops
    )

    assert torch.all(edge_index_from_shape == edge_index_from_ptr)


@pytest.mark.parametrize("B, N", [(1, 5), (4, 9)])
def test_ptr_batch_invertibility(B, N):
    ptr = torch.arange(B + 1) * N
    batch = get_batch_from_ptr(ptr)
    ptr_test = get_ptr_from_batch(batch)
    assert torch.all(ptr == ptr_test)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

project = "PELICAN-lite"
copyright = "2025, Jonas Spinner"
author = "Jonas Spinner"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",
]
autosummary_generate = True
autoclass_content = "both"
autodoc_class_signature = "mixed"

autodoc_default_options = {
    "members": True,
    "inherited-members": False,
}
templates_path = ["_templates"]
napoleon_custom_sections = [("params_style")]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



import math

import pytest

from pelican.layers import (
    Aggregator0to2,
    Aggregator1to2,
    Aggregator2to0,
    Aggregator2to1,
    Aggregator2to2,
    PELICANBlock,
)

from .utils import generate_batch


@pytest.mark.parametrize("aggr", ["sum", "prod", "mean", "amax", "amin"])
@pytest.mark.parametrize("map_multipliers", [True, False])
@pytest.mark.parametrize("factorize", [True, False])
@pytest.mark.parametrize(
    "aggregator,in_rank,out_rank",
    [
        [Aggregator0to2, 0, 2],
        [Aggregator1to2, 1, 2],
        [Aggregator2to0, 2, 0],
        [Aggregator2to1, 2, 1],
        [Aggregator2to2, 2, 2],
    ],
)
@pytest.mark.parametrize("in_channels,out_channels", [(16, 16), (7, 13)])
def test_shape_aggregator(
    aggregator,
    in_rank,
    out_rank,
    in_channels,
    out_channels,
    map_multipliers,
    factorize,
    aggr,
):
    batch, edge_index, graph, nodes, edges = generate_batch(C=in_channels)
    G = batch[-1].item() + 1
    N = batch.size(0)
    E = edge_index.size(1)

    in_data = {0: graph, 1: nodes, 2: edges}[in_rank]
    out_objs = {0: G, 1: N, 2: E}[out_rank]

    handler = aggregator(
        in_channels=in_channels,
        out_channels=out_channels,
        map_multipliers=map_multipliers,
        factorize=factorize,
        aggr=aggr,
    )
    out = handler(in_data, edge_index, batch, G=G)
    assert out.shape == (out_objs, out_channels)


@pytest.mark.parametrize("aggr", ["sum", "prod", "mean", "amax", "amin"])
@pytest.mark.parametrize("activation", ["relu", "gelu", "leaky_relu", "tanh", "sigmoid", "silu"])
@pytest.mark.parametrize("map_multipliers", [True, False])
@pytest.mark.parametrize("factorize", [True, False])
@pytest.mark.parametrize(
    "hidden_channels,increase_hidden_channels", [(16, 1), (13, math.pi), (19, 0.123)]
)
def test_shape_block(
    map_multipliers,
    factorize,
    aggr,
    activation,
    hidden_channels,
    increase_hidden_channels,
):
    batch, edge_index, _, _, edges = generate_batch(C=hidden_channels)
    E = edge_index.size(1)

    handler = PELICANBlock(
        hidden_channels=hidden_channels,
        increase_hidden_channels=increase_hidden_channels,
        activation=activation,
        map_multipliers=map_multipliers,
        factorize=factorize,
        aggr=aggr,
    )
    out = handler(edges, edge_index, batch)
    assert out.shape == (E, hidden_channels)



import math

import pytest
import torch

from pelican.nets import PELICAN

from .utils import generate_batch, permute_single_graph


def run_shape_test(
    num_blocks,
    hidden_channels,
    increase_hidden_channels,
    in_channels_rank0,
    in_channels_rank1,
    in_channels_rank2,
    out_rank,
    out_channels,
    map_multipliers,
    factorize,
    checkpoint_blocks=False,
    give_G=True,
    compile=False,
):
    batch, edge_index, graph, _, _ = generate_batch(C=in_channels_rank0)
    _, _, _, nodes, _ = generate_batch(C=in_channels_rank1, batch=batch, edge_index=edge_index)
    _, _, _, _, edges = generate_batch(C=in_channels_rank2, batch=batch, edge_index=edge_index)
    G = batch[-1].item() + 1
    N = batch.size(0)
    E = edge_index.size(1)
    out_objs = {0: G, 1: N, 2: E}[out_rank]

    net = PELICAN(
        num_blocks=num_blocks,
        hidden_channels=hidden_channels,
        increase_hidden_channels=increase_hidden_channels,
        in_channels_rank0=in_channels_rank0,
        in_channels_rank1=in_channels_rank1,
        in_channels_rank2=in_channels_rank2,
        out_rank=out_rank,
        out_channels=out_channels,
        map_multipliers=map_multipliers,
        factorize=factorize,
        compile=compile,
        checkpoint_blocks=checkpoint_blocks,
    )
    out = net(
        in_rank2=edges,
        in_rank1=nodes,
        in_rank0=graph,
        batch=batch,
        edge_index=edge_index,
        num_graphs=G if give_G else None,
    )
    assert out.shape == (out_objs, out_channels)


@pytest.mark.parametrize("hidden_channels,increase_hidden_channels", [(16, 1), (7, math.pi)])
@pytest.mark.parametrize("num_blocks", [0, 1, 3])
@pytest.mark.parametrize(
    "in_channels_rank0,in_channels_rank1,in_channels_rank2",
    [(0, 0, 1), (1, 0, 0), (0, 1, 0), (1, 1, 1)],
)
@pytest.mark.parametrize("out_rank,out_channels", [(0, 1), (1, 2), (2, 3)])
@pytest.mark.parametrize("map_multipliers", [True, False])
@pytest.mark.parametrize("factorize", [True, False])
@pytest.mark.parametrize("checkpoint_blocks", [False, True])
@pytest.mark.parametrize("give_G", [False, True])
def test_shape(
    num_blocks,
    hidden_channels,
    increase_hidden_channels,
    in_channels_rank0,
    in_channels_rank1,
    in_channels_rank2,
    out_rank,
    out_channels,
    map_multipliers,
    factorize,
    checkpoint_blocks,
    give_G,
    compile=False,
):
    run_shape_test(
        num_blocks=num_blocks,
        hidden_channels=hidden_channels,
        increase_hidden_channels=increase_hidden_channels,
        in_channels_rank0=in_channels_rank0,
        in_channels_rank1=in_channels_rank1,
        in_channels_rank2=in_channels_rank2,
        out_rank=out_rank,
        out_channels=out_channels,
        map_multipliers=map_multipliers,
        factorize=factorize,
        compile=compile,
        checkpoint_blocks=checkpoint_blocks,
        give_G=give_G,
    )


@pytest.mark.parametrize("hidden_channels,increase_hidden_channels", [(16, 1), (7, math.pi)])
@pytest.mark.parametrize("num_blocks", [0, 1, 3])
@pytest.mark.parametrize(
    "in_channels_rank0,in_channels_rank1,in_channels_rank2",
    [(0, 0, 1), (1, 0, 0), (0, 1, 0), (1, 1, 1)],
)
@pytest.mark.parametrize("out_rank,out_channels", [(0, 1), (1, 2), (2, 3)])
@pytest.mark.parametrize("map_multipliers", [True, False])
@pytest.mark.parametrize("factorize", [True, False])
def test_permutation_equivariance(
    num_blocks,
    hidden_channels,
    increase_hidden_channels,
    in_channels_rank0,
    in_channels_rank1,
    in_channels_rank2,
    out_rank,
    out_channels,
    map_multipliers,
    factorize,
):
    # only test permutation equivariance on single graphs for simplicity
    batch, edge_index, graph, _, _ = generate_batch(C=in_channels_rank0, G=1)
    _, _, _, nodes, _ = generate_batch(C=in_channels_rank1, batch=batch, edge_index=edge_index)
    _, _, _, _, edges = generate_batch(C=in_channels_rank2, batch=batch, edge_index=edge_index)
    G = batch[-1].item() + 1

    net = PELICAN(
        num_blocks=num_blocks,
        hidden_channels=hidden_channels,
        increase_hidden_channels=increase_hidden_channels,
        in_channels_rank0=in_channels_rank0,
        in_channels_rank1=in_channels_rank1,
        in_channels_rank2=in_channels_rank2,
        out_rank=out_rank,
        out_channels=out_channels,
        map_multipliers=map_multipliers,
        factorize=factorize,
    )

    # path 1: first aggregate, then permute
    out = net(
        in_rank2=edges,
        in_rank1=nodes,
        in_rank0=graph,
        batch=batch,
        edge_index=edge_index,
        num_graphs=G,
    )
    kwargs = {
        0: {"graph": out, "nodes": None, "edges": None},
        1: {"graph": None, "nodes": out, "edges": None},
        2: {"graph": None, "nodes": None, "edges": out},
    }[out_rank]
    (
        permutation,
        graph_perm,
        nodes_perm,
        edges_perm,
        edge_index_perm,
    ) = permute_single_graph(**kwargs, edge_index=edge_index)
    out_permlater = {0: graph_perm, 1: nodes_perm, 2: edges_perm}[out_rank]

    # path 2: first permute, then aggregate
    _, graph_perm, nodes_perm, edges_perm, _ = permute_single_graph(
        permutation=permutation,
        graph=graph,
        nodes=nodes,
        edges=edges,
        edge_index=edge_index,
    )
    out_permfirst = net(
        in_rank2=edges_perm,
        in_rank1=nodes_perm,
        in_rank0=graph_perm,
        batch=batch,
        edge_index=edge_index_perm,
        num_graphs=G,
    )

    torch.testing.assert_close(out_permlater, out_permfirst)


def test_compile(
    num_blocks=1,
    hidden_channels=16,
    increase_hidden_channels=1,
    in_channels_rank0=1,
    in_channels_rank1=1,
    in_channels_rank2=1,
    out_rank=2,
    out_channels=3,
    map_multipliers=True,
    factorize=False,
):
    run_shape_test(
        num_blocks=num_blocks,
        hidden_channels=hidden_channels,
        increase_hidden_channels=increase_hidden_channels,
        in_channels_rank0=in_channels_rank0,
        in_channels_rank1=in_channels_rank1,
        in_channels_rank2=in_channels_rank2,
        out_rank=out_rank,
        out_channels=out_channels,
        map_multipliers=map_multipliers,
        factorize=factorize,
        compile=True,
    )



"""Layers for PELICAN architecture."""

import torch
from torch import nn

from .primitives import (
    aggregate_0to2,
    aggregate_1to2,
    aggregate_2to0,
    aggregate_2to1,
    aggregate_2to2,
    bell_number,
)

ACTIVATION = {
    "leaky_relu": nn.LeakyReLU(),
    "relu": nn.ReLU(),
    "gelu": nn.GELU(),
    "tanh": nn.Tanh(),
    "sigmoid": nn.Sigmoid(),
    "silu": nn.SiLU(),
}


class GeneralAggregator(nn.Module):
    """General aggregator class."""

    def __init__(
        self,
        aggregator: callable,
        in_rank: int,
        out_rank: int,
        in_channels: int,
        out_channels: int,
        map_multipliers: bool = True,
        factorize: bool = False,
        aggr: str = "mean",
    ):
        """
        Parameters
        ----------
        aggregator : callable
            Aggregation function to use.
        in_rank : int
            Input rank (0 for graph, 1 for nodes, 2 for edges).
        out_rank : int
            Output rank (0 for graph, 1 for nodes, 2 for edges).
        in_channels : int
            Number of input channels.
        out_channels : int
            Number of output channels.
        map_multipliers : bool
            Whether to use learnable multipliers for each aggregation map, by default True.
        factorize : bool
            Whether to use factorized coefficients, by default False.
            Factorization reduces the number of parameters.
        aggr : str
            Aggregation method to use ('mean', 'sum', 'prod', 'amin', 'amax'), by default 'mean'.
        """
        super().__init__()
        self.num_maps = bell_number(in_rank + out_rank)
        self.in_channels = in_channels

        self.aggr = aggr
        self.aggregator = aggregator
        self.factorize = factorize

        self.map_multipliers = nn.Parameter(torch.ones(self.num_maps)) if map_multipliers else None

        if factorize:
            self.coeffs00 = nn.Parameter(torch.empty(in_channels, self.num_maps))
            self.coeffs01 = nn.Parameter(torch.empty(out_channels, self.num_maps))
            self.coeffs10 = nn.Parameter(torch.empty(in_channels, out_channels))
            self.coeffs11 = nn.Parameter(torch.empty(in_channels, out_channels))
            nn.init.kaiming_uniform_(self.coeffs00, nonlinearity="linear")
            nn.init.kaiming_uniform_(self.coeffs01, nonlinearity="linear")
            nn.init.kaiming_uniform_(self.coeffs10, nonlinearity="linear")
            nn.init.kaiming_uniform_(self.coeffs11, nonlinearity="linear")
        else:
            self.coeffs_direct = nn.Parameter(torch.empty(in_channels, out_channels, self.num_maps))
            nn.init.kaiming_uniform_(self.coeffs_direct, nonlinearity="relu")

    @property
    def coeffs(self):
        if self.factorize:
            coeffs = self.coeffs00.unsqueeze(1) * self.coeffs10.unsqueeze(
                2
            ) + self.coeffs01.unsqueeze(0) * self.coeffs11.unsqueeze(2)
        else:
            coeffs = self.coeffs_direct
        if self.map_multipliers is not None:
            coeffs = coeffs * self.map_multipliers.view(1, 1, self.num_maps)
        return coeffs

    def forward(self, x, *args, **kwargs):
        """Forward pass of the aggregator.

        Parameters
        ----------
        x : torch.Tensor
            Input features of shape (in_objects, in_channels).
        *args
            Additional arguments to pass to the aggregator.
        **kwargs
            Additional keyword arguments to pass to the aggregator.

        Returns
        -------
        out : torch.Tensor
            Output features of shape (out_objects, in_channels).
        """
        x = self.aggregator(x, *args, reduce=self.aggr, **kwargs)

        in_objects, in_channels, num_maps = x.shape
        x_flat = x.reshape(in_objects, in_channels * num_maps)
        coeffs_flat = self.coeffs.reshape(-1, in_channels * num_maps).contiguous()
        out = torch.nn.functional.linear(x_flat, coeffs_flat)
        return out


class Aggregator2to2(GeneralAggregator):
    """Aggregator from edges (rank 2) to edges (rank 2)."""

    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(aggregate_2to2, 2, 2, in_channels, out_channels, **kwargs)


class Aggregator2to1(GeneralAggregator):
    """Aggregator from edges (rank 2) to nodes (rank 1)."""

    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(aggregate_2to1, 2, 1, in_channels, out_channels, **kwargs)


class Aggregator2to0(GeneralAggregator):
    """Aggregator from edges (rank 2) to graph (rank 0)."""

    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(aggregate_2to0, 2, 0, in_channels, out_channels, **kwargs)


class Aggregator1to2(GeneralAggregator):
    """Aggregator from nodes (rank 1) to edges (rank 2)."""

    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(aggregate_1to2, 1, 2, in_channels, out_channels, **kwargs)


class Aggregator0to2(GeneralAggregator):
    """Aggregator from graph (rank 0) to edges (rank 2)."""

    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(aggregate_0to2, 0, 2, in_channels, out_channels, **kwargs)


class PELICANBlock(nn.Module):
    """PELICAN edge-to-edge aggregation block.

    A single PELICAN block consisting of a feedforward network and edge-to-edge aggregation."""

    def __init__(
        self,
        hidden_channels: int,
        increase_hidden_channels: float = 1.0,
        activation: str = "leaky_relu",
        dropout_prob: float | None = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        hidden_channels : int
            Number of hidden channels.
        increase_hidden_channels : float
            Factor to increase hidden channels in the feedforward network, by default 1.0.
        activation : str
            Activation function to use ('gelu', 'relu', 'leaky_relu', 'tanh', 'sigmoid', 'silu'), by default 'leaky_relu'.
        dropout_prob: float
            Dropout probability, by default None.
        **kwargs
            Additional keyword arguments to pass to the aggregator.
        """
        super().__init__()
        hidden_channels_2 = int(increase_hidden_channels * hidden_channels)
        linear_in = nn.Linear(hidden_channels, hidden_channels_2)
        self.activation = ACTIVATION[activation]
        norm = nn.RMSNorm(normalized_shape=hidden_channels_2)
        dropout = nn.Dropout(p=dropout_prob) if dropout_prob is not None else nn.Identity()
        self.mlp = nn.ModuleList([linear_in, self.activation, norm, dropout])

        self.aggregator = Aggregator2to2(
            in_channels=hidden_channels_2,
            out_channels=hidden_channels,
            **kwargs,
        )

    def forward(self, x, edge_index, batch, **kwargs):
        """Forward pass of the PELICAN block.

        Parameters
        ----------
        x : torch.Tensor
            Edge-level features of shape (E, C).
        edge_index : torch.Tensor
            Edge index tensor of shape (2, E).
        batch : torch.Tensor
            Batch tensor of shape (N).
        **kwargs

        Returns
        -------
        x : torch.Tensor
            Updated edge-level features of shape (E, C).
        """
        for layer in self.mlp:
            x = layer(x)

        x = self.aggregator(x, edge_index=edge_index, batch=batch, **kwargs)
        x = self.activation(x)
        return x



"""Primitives for aggregating features between different ranks in a graph."""

import torch


def bell_number(n: int) -> int:
    """Compute the Bell number for small n (0 <= n <= 4).
    The Bell number is the number of ways to partition a set of n elements.
    In PELICAN, it corresponds to the number of aggregation maps
    for a given rank n = in_rank + out_rank.
    """
    if n == 0:
        return 1
    elif n == 1:
        return 1
    elif n == 2:
        return 2
    elif n == 3:
        return 5
    elif n == 4:
        return 15
    else:
        raise NotImplementedError(
            f"Asking for bell_number(n={n}), but bell_number(n>4) not implemented yet."
        )


def aggregate_0to2(graph, edge_index, batch, **kwargs):
    """Aggregate from graph (rank 0) to edges (rank 2).

    Parameters
    ----------
    graph : torch.Tensor
        Graph-level features of shape (G, C).
    edge_index : torch.Tensor
        Edge index tensor of shape (2, E).
    batch : torch.Tensor
        Batch tensor of shape (N).
    **kwargs

    Returns
    -------
    ops : torch.Tensor
        Aggregated features of shape (E, C, 2).
    """
    row, col = edge_index
    edge_batch = batch[row]
    is_diag = row == col
    diag_mask = is_diag.unsqueeze(-1).type_as(graph)

    ops = torch.stack(
        [
            graph[edge_batch],
            graph[edge_batch] * diag_mask,
        ],
        dim=-1,
    )  # shape (E, C, 2)
    return ops


def aggregate_1to2(nodes, edge_index, batch, reduce: str = "mean", **kwargs):
    """Aggregate from nodes (rank 1) to edges (rank 2).

    Parameters
    ----------
    nodes : torch.Tensor
        Node-level features of shape (N, C).
    edge_index : torch.Tensor
        Edge index tensor of shape (2, E).
    batch : torch.Tensor
        Batch tensor of shape (N).
    reduce : str, optional
        Reduction method to use ('mean', 'sum', 'prod', 'amin', 'amax'), by default 'mean'.
    **kwargs

    Returns
    -------
    ops : torch.Tensor
        Aggregated features of shape (E, C, 5).
    """
    _, C = nodes.shape
    E = edge_index.size(1)
    row, col = edge_index
    edge_batch = batch[row]
    is_diag = row == col
    diag_mask = is_diag.unsqueeze(-1).type_as(nodes)

    nodes_agg = custom_scatter(nodes, batch, dim_size=E, C=C, reduce=reduce)
    is_diag = is_diag.unsqueeze(-1)

    ops = torch.stack(
        [
            nodes[row] * diag_mask,
            nodes[row],
            nodes[col],
            nodes_agg[edge_batch] * diag_mask,
            nodes_agg[edge_batch],
        ],
        dim=-1,
    )  # shape (E, C, 5)
    return ops


def aggregate_2to0(edges, edge_index, batch, reduce: str = "mean", G=None, **kwargs):
    """Aggregate from edges (rank 2) to graph (rank 0).

    Parameters
    ----------
    edges : torch.Tensor
        Edge-level features of shape (E, C).
    edge_index : torch.Tensor
        Edge index tensor of shape (2, E).
    batch : torch.Tensor
        Batch tensor of shape (N).
    reduce : str, optional
        Reduction method to use ('mean', 'sum', 'prod', 'amin', 'amax'), by default 'mean'.
    G : int, optional
        Number of graphs in the batch. If None, it will be inferred from the batch tensor.
    **kwargs

    Returns
    -------
    ops : torch.Tensor
        Aggregated features of shape (G, C, 2).
    """
    _, C = edges.shape
    if G is None:
        # host synchronization causes slowdown; maybe there is a better way?
        G = batch[-1].item() + 1
    row, col = edge_index
    edge_batch = batch[row]
    is_diag = row == col
    diag_mask = is_diag.unsqueeze(-1).type_as(edges)

    diags = edges * diag_mask
    graph_agg = custom_scatter(edges, edge_batch, dim_size=G, C=C, reduce=reduce)
    diag_agg = custom_scatter(diags, edge_batch, dim_size=G, C=C, reduce=reduce)

    ops = torch.stack(
        [
            graph_agg,
            diag_agg,
        ],
        dim=-1,
    )  # shape (G, C, 2)
    return ops


def aggregate_2to1(edges, edge_index, batch, reduce: str = "mean", **kwargs):
    """Aggregate from edges (rank 2) to nodes (rank 1).

    Parameters
    ----------
    edges : torch.Tensor
        Edge-level features of shape (E, C).
    edge_index : torch.Tensor
        Edge index tensor of shape (2, E).
    batch : torch.Tensor
        Batch tensor of shape (N).
    reduce : str, optional
        Reduction method to use ('mean', 'sum', 'prod', 'amin', 'amax'), by default 'mean'.
    **kwargs

    Returns
    -------
    ops : torch.Tensor
        Aggregated features of shape (N, C, 5).
    """
    _, C = edges.shape
    N = batch.size(0)
    row, col = edge_index
    edge_batch = batch[row]
    is_diag = row == col
    diag_mask = is_diag.unsqueeze(-1).type_as(edges)

    diags_padded = edges * diag_mask
    diags = custom_scatter(diags_padded, row, dim_size=N, C=C, reduce="sum")
    row_agg = custom_scatter(edges, row, dim_size=N, C=C, reduce=reduce)
    col_agg = custom_scatter(edges, col, dim_size=N, C=C, reduce=reduce)
    graph_agg = custom_scatter(edges, edge_batch, dim_size=N, C=C, reduce=reduce)
    diag_agg = custom_scatter(diags_padded, edge_batch, dim_size=N, C=C, reduce=reduce)

    ops = torch.stack(
        [
            diags,
            row_agg,
            col_agg,
            graph_agg[batch],
            diag_agg[batch],
        ],
        dim=-1,
    )  # shape (N, C, 5)
    return ops


def aggregate_2to2(edges, edge_index, batch, reduce: str = "mean", **kwargs):
    """Aggregate from edges (rank 2) to edges (rank 2).

    Parameters
    ----------
    edges : torch.Tensor
        Edge-level features of shape (E, C).
    edge_index : torch.Tensor
         Edge index tensor of shape (2, E).
    batch : torch.Tensor
        Batch tensor of shape (N).
    reduce : str, optional
        Reduction method to use ('mean', 'sum', 'prod', 'amin', 'amax'), by default 'mean'.
    **kwargs

    Returns
    -------
    ops : torch.Tensor
        Aggregated features of shape (E, C, 15).
    """
    _, C = edges.shape
    N = batch.size(0)
    row, col = edge_index
    perm_T = get_transpose(edge_index)
    edge_batch = batch[row]
    is_diag = row == col
    diag_mask = is_diag.unsqueeze(-1).type_as(edges)

    diags_padded = edges * diag_mask
    diags = custom_scatter(diags_padded, row, dim_size=N, C=C, reduce="sum")
    row_agg = custom_scatter(edges, row, dim_size=N, C=C, reduce=reduce)
    col_agg = custom_scatter(edges, col, dim_size=N, C=C, reduce=reduce)
    graph_agg = custom_scatter(edges, edge_batch, dim_size=N, C=C, reduce=reduce)
    diag_agg = custom_scatter(diags_padded, edge_batch, dim_size=N, C=C, reduce=reduce)

    # Note: creating ops as empty tensor and filling it is ~4% faster but requires ~10% more RAM
    ops = torch.stack(
        [
            edges,
            edges[perm_T],
            diags_padded,
            diags[row],
            diags[col],
            col_agg[row] * diag_mask,
            row_agg[col] * diag_mask,
            col_agg[row],
            col_agg[col],
            row_agg[row],
            row_agg[col],
            diag_agg[edge_batch],
            diag_agg[edge_batch] * diag_mask,
            graph_agg[edge_batch],
            graph_agg[edge_batch] * diag_mask,
        ],
        dim=-1,
    )  # shape (E, C, 15)
    return ops


def get_transpose(edge_index):
    """Get the permutation that transposes the edge index."""
    row, col = edge_index
    key = (row << 32) | col
    rev_key = (col << 32) | row
    key_sorted, perm = key.sort()
    idx = torch.searchsorted(key_sorted, rev_key)
    return perm[idx]


def custom_scatter(src, index, dim_size: int, C: int, reduce: "str"):
    """Custom scatter function to conveniently aggregate features, falling back to scatter_reduce_."""
    out = src.new_zeros(dim_size, C)
    out.scatter_reduce_(
        0, index.unsqueeze(-1).expand(-1, C), src, reduce=reduce, include_self=False
    )
    return out



import pytest
import torch

from pelican.primitives import (
    aggregate_0to2,
    aggregate_1to2,
    aggregate_2to0,
    aggregate_2to1,
    aggregate_2to2,
)

from .utils import generate_batch, permute_single_graph


@pytest.mark.parametrize("reduce", ["sum", "prod", "mean", "amax", "amin"])
@pytest.mark.parametrize(
    "aggregator,in_rank,out_rank",
    [
        [aggregate_0to2, 0, 2],
        [aggregate_1to2, 1, 2],
        [aggregate_2to0, 2, 0],
        [aggregate_2to1, 2, 1],
        [aggregate_2to2, 2, 2],
    ],
)
def test_shape(aggregator, in_rank, out_rank, reduce):
    batch, edge_index, graph, nodes, edges = generate_batch()
    G = batch[-1].item() + 1
    N = batch.size(0)
    E = edge_index.size(1)
    C = graph.size(1)

    in_data = {0: graph, 1: nodes, 2: edges}[in_rank]
    out_objs = {0: G, 1: N, 2: E}[out_rank]

    out = aggregator(in_data, edge_index, batch, reduce=reduce, G=G)
    assert out.shape[:2] == (out_objs, C)


@pytest.mark.parametrize("reduce", ["sum", "prod", "mean", "amax", "amin"])
@pytest.mark.parametrize(
    "aggregator,in_rank,out_rank",
    [
        [aggregate_0to2, 0, 2],
        [aggregate_1to2, 1, 2],
        [aggregate_2to0, 2, 0],
        [aggregate_2to1, 2, 1],
        [aggregate_2to2, 2, 2],
    ],
)
def test_permutation_equivariance(aggregator, in_rank, out_rank, reduce):
    # only test permutation equivariance on single graphs for simplicity
    batch, edge_index, graph, nodes, edges = generate_batch(G=1, N_min=2, N_max=3, C=1)
    G = batch[-1].item() + 1

    # path 1: first aggregate, then permute
    in_data = {0: graph, 1: nodes, 2: edges}[in_rank]
    out = aggregator(in_data, edge_index, batch, reduce=reduce, G=G)
    kwargs = {
        0: {"graph": out, "nodes": None, "edges": None},
        1: {"graph": None, "nodes": out, "edges": None},
        2: {"graph": None, "nodes": None, "edges": out},
    }[out_rank]
    (
        permutation,
        graph_perm,
        nodes_perm,
        edges_perm,
        edge_index_perm,
    ) = permute_single_graph(**kwargs, edge_index=edge_index)
    out_permlater = {0: graph_perm, 1: nodes_perm, 2: edges_perm}[out_rank]

    # path 2: first permute, then aggregate
    _, graph_perm, nodes_perm, edges_perm, _ = permute_single_graph(
        graph=graph,
        nodes=nodes,
        edges=edges,
        edge_index=edge_index,
        permutation=permutation,
    )
    in_data_perm = {0: graph_perm, 1: nodes_perm, 2: edges_perm}[in_rank]
    out_permfirst = aggregator(in_data_perm, edge_index_perm, batch, reduce=reduce, G=G)

    torch.testing.assert_close(out_permlater, out_permfirst)
